#!/usr/bin/env python3

import xml.etree.ElementTree as ET
from nltk.tokenize import RegexpTokenizer
from nltk.corpus import stopwords
from nltk.metrics import edit_distance
from pyxdameraulevenshtein import damerau_levenshtein_distance_ndarray, normalized_damerau_levenshtein_distance_ndarray
import numpy as np
import pandas as pd
import string
import re


































# functions from hf_extractdiscussions.property
# ==============================================================================

# function to exclude overlapping textboxes between documents
# input:
# - dict_text: dictionary of texts of one document
# - dict_overlaps_year: dictionary with overlaps
# output:
# - dict_text: modified dict_text
def exclude_overlaps(dict_text, dict_overlaps):
    # initialize to impossible values
    first_entry = -1
    last_entry = 1000

    # get index of textbox from first and last page
    # the overlap dictionary only contains an entry, if an overlap was detected
    for entry, array in dict_overlaps.items():
        if entry == 'first':
            first_entry = int(array[0])
        if entry == 'last':
            last_entry = int(array[0])

    # get list of keys for first and last page
    list_first_page = [key for key in dict_text if key.split(',')[1] == '0']
    last_page = max([int(key.split(',')[1]) for key in dict_text])
    list_last_page = [key for key in dict_text if key.split(',')[1] == str(last_page)]

    # modify dict_text on first page
    for key in list_first_page:
        if int(key.split(',')[2]) < first_entry:
            dict_text[key] = ''

    # ... and on last page
    for key in list_last_page:
        if int(key.split(',')[2]) > last_entry:
            dict_text[key] = ''

    return dict_text


# tokenizer
tokenizer_canton = RegexpTokenizer(r'\w+')    # only leaves words
#tokenizer = RegexpTokenizer(r'\w+(?:-\w+)*|\$[\d\.]+|\S+')
# last part \S+ is needed to get colon, \S stands for white space
tokenizer = RegexpTokenizer(r'\w+(?:-/w+)*|\$[\d\.]+')

# function to extract discussion starts
# !!! maybe we only need a list of discussion starts
# input:
# - dict_text: dictionary with text of one file
# - list_names: list of MPs
# - list_stopwords: list of german and french stopwords
# - bln_print: whether to print during execution, default False
# output:
# - dict_discussionstarts: dictionary with discussion starts
def get_discussion_starts(dict_text, df_names, list_stopwords, bln_print=False):

    # initialize empty dictionary
    dict_discussionstarts = {}

    # add a few terms to list_stopwords that are easily mistaken as last names
    list_stopwords.extend(['art', 'rath', 'alinea', 'stimmen', 'stimme', 'hans', 'walter', 'werner', 'projet', 'stimmt', 'nicht', 'vote', 'pas', 'gallen', 'stgallen',
                           'kasse', 'fasse', 'sitten', 'herren', 'herr', 'alter'])

    # for every textbox
    for key, text in dict_text.items():
        if ':' in text[:100]:
            # extract the index of the colon in the text
            colon_index_text = text.index(':')

            # look at first few terms of that textbox
            text_start = re.sub(r'[\(\)]','',text[:colon_index_text])
            print('text_start', text_start)
            list_oi = tokenizer.tokenize(text_start)
            print('asdf', list_oi)

            # shorten to part before colon
            list_oi2 = list_oi

            # remove stopwords
            list_oi2 = [term for term in list_oi2 if term.lower() not in list_stopwords]

            # remove punctuation
            list_oi2 = [''.join(c for c in s if c not in string.punctuation) for s in list_oi2]
            list_oi2 = [s for s in list_oi2 if s]

            # remove lower case terms
#            list_oi2 = [term for term in list_oi2 if not term.islower()]

            # remove numbers
            list_oi2 = [term for term in list_oi2 if not term.isdigit()]

            # remove single characters
            list_oi2 = [term for term in list_oi2 if len(term)>1]

            # initialize string for name and role
            str_name = ''
            str_role = ''
            int_uniqueID = int(0)
            str_canton = ''

            # for every term, reversed finds canton before it finds name
            for term in reversed(list_oi2):
                # if possible, find a name in a list
                str_name, str_role, int_uniqueID, str_canton = find_names(term, str_name, str_role, int_uniqueID, df_names, str_canton, bln_print=True)
            print('name', str_name, 'role', str_role)

            # get rid of 'Präsident stimmt nicht Président ne vote pas'
            if set(str_role.split()).intersection(set(['Präsident', 'Präsidentin', 'Président', 'Présidente'])) and not str_name:
                print('++++++++++ Präsident', list_oi2, list_oi)
                if set(['stimmt', 'nicht', 'vote', 'pas']).intersection(list_oi):
                    str_role = ''

            # get rid of 'Für den Antrag "Name" stimmen: Votent pour la proposition "Name":'
            if str_name:
                print('++++++++++ Name', list_oi2, list_oi)
                if len(set(['Antrag', 'stimmen', 'Votent', 'proposition']).intersection(list_oi)) > 1:
                    str_name = ''

            # if a name has been found
            if str_name or str_role:
                # add text to dictionary
                dict_discussionstarts[(key, str_name, str_role, int_uniqueID)] = (text[:colon_index_text],
                        text[colon_index_text+1:])
                if bln_print:
                    print('found a name:', list_oi2, str_name, str_role, '\n')

    return dict_discussionstarts



# small function to get first item of tupels in a list
def get_first_item(list_tupels):
    list_first_item = [tupel[0] for tupel in list_tupels]
    return list_first_item

# small function to get last two items of tupels in a list
def get_last_item(list_tupels):
    list_last_item = [tupel[-2:] for tupel in list_tupels]
    return list_last_item



# function to find names
# input:
# - term: term that might be name
# - str_name: string to which name should be attached
# - str_role: string to which role should be attached
# - int_uniqueID: integer for uniqueID
#                 !!! (if there are several possibilities, this becomes a tuple)
# - list_tupels: list of tupels containing all types of names
def find_names(term, str_name, str_role, int_uniqueID, df_names, str_canton, bln_print=False):

    def get_string(term, str_name, str_role, int_uniqueID, str_canton):
        name_type = ''
        # if it is one of the simple names
        if term in list(df_names['name_short'].loc[df_names['type']=='simple']):
            str_name = add_to_string(str_name, term)
            name_type = 'simple'
        # if it is a role
        elif term in list_roles:
            str_role = add_to_string(str_role, term)
        # if it is a double name
        elif term in list(df_names['name_short'].loc[df_names['type']=='double']):
            print(20*'\n', 'DOUBLE NAME')
            # get correct name
            correct_name = df_names.loc[(df_names['type']=='double') & (df_names['name_short']== term)].iat[0, df_names.columns.get_loc('name_correct')]
            if bln_print:
                print('double name', correct_name)
            str_name = add_to_string(str_name, correct_name)
            name_type = 'double'
        # if it is a composite name
        elif term in list(df_names['name_short'].loc[df_names['type']=='comp']):
            # get correct name
            correct_name = df_names.loc[(df_names['type']=='comp') & (df_names['name_short']== term)].iat[0, df_names.columns.get_loc('name_correct')]
            if bln_print:
                print('composite name', correct_name)
            str_name = add_to_string(str_name, correct_name)
            name_type = 'comp'
        # if it contains a canton
        # !!! also pass list_oi to look for canton
        # !!! how to handle for people mentioned in text???
        elif term in list(df_names['name_short'].loc[df_names['type']=='canton']):
            if bln_print:
                print('contains a canton', term)
#            canton_missing = False
#            df_temp = df_names.loc[df_names['name_short']==term]
#            print('list_correct', df_temp)
#            print(str_canton)
#            if str_canton:
#                str_correct = check_place(df_temp, str_canton)
#                if str_correct in ['not found', 'too many']:
#                    str_name = add_to_string(str_name, term + ' (CANTONT MISSING)')
#                    canton_missing = True
#                else:
#                    str_name = add_to_string(str_name, str_temp)
#                    name_type = 'canton'
            str_name = add_to_string(str_name, term + ' (CANTON MISSING)')
            name_type = 'canton'


        temp = ''
        if name_type in ['simple', 'double', 'comp']:
            temp = df_names.loc[(df_names['type']==name_type) & (df_names['name_short']==term)].iat[0, df_names.columns.get_loc('uniqueIndex')]
        elif name_type in ['canton']:
            temp = tuple(df_names.loc[(df_names['type']==name_type) & (df_names['name_short']==term)].iloc[:, df_names.columns.get_loc('uniqueIndex')])
#            if canton_missing:
#                temp = tuple(df_names.loc[(df_names['type']==name_type) & (df_names['name_short']==term)].iloc[:, df_names.columns.get_loc('uniqueIndex')])
#            else:
#                temp = df_names.loc[(df_names['type']==name_type) & (df_names['name_correct']==str_correct)].iat[0, df_names.columns.get_loc('uniqueIndex')]

        if temp:
            if int_uniqueID == 0:
                int_uniqueID = temp
            else:
                int_uniqueID = (int_uniqueID, temp)

        return str_name, str_role, int_uniqueID

    # small function to add term to str_name
    def add_to_string(string, term):
        if not string:
            string = term
        else:
            string += ' ' + term
        return string

    list_roles = ['Präsident', 'Präsidentin', 'Vizepräsident', 'Präsidium', 'Président', 'Présidente', 'président', 'présidente',
                  'Berichterstatter', 'Berichterstatterin', 'rapporteur',
                  'Sprecher', 'Sprecherin', 'porte-parole', 'porteparole',
                  'Bundesrat', 'Bundesrath', 'Bundesrätin', 'conseiller fédéral',
                  'Vizepräsident']

    list_notnames = ['Art', 'Rath', 'Alinea', 'Stimmen', 'Stimme', 'Hans', 'Walter', 'Werner', 'projet', 'stimmt', 'nicht', 'vote', 'pas', 'Gallen', 'StGallen',
                     'Kasse', 'fasse', 'Sitten', 'Herren', 'Herr', 'Alter']

    list_places = get_list_cantons(df_names)

    if bln_print:
        print('now is about: ------', term)
    # extract list and array of last names
    list_all_names = list(df_names['name_short'])
    array_all_names = np.array(df_names['name_short'])

    # if term is not easily mistaken as a name (avoid false positives)
    if term not in list_notnames:

        # if term is in the list of all names and roles
        if term in (list_all_names + list_roles):
            # get correct name and uniqueID, or role, for that term
            str_name, str_role, int_uniqueID = get_string(term, str_name, str_role, int_uniqueID, str_canton)

            if bln_print:
                print('=== correct name', term)
        # if term in list of cantons
        elif term in list_places:
            str_canton = term
        # if term is not in list_all_names
        else:
            # look for similar names based on (normalized) Damerau-Levenshtein distance
            # !!! probably need to improve this procedure
            #       - find better values ....
            if bln_print:
                print(term)
            array_normalized = array_all_names[normalized_damerau_levenshtein_distance_ndarray(term, array_all_names) <= 0.35]
            array_normalized_values = normalized_damerau_levenshtein_distance_ndarray(term, array_normalized)
            if bln_print:
                print(array_normalized, array_normalized_values)
            array_absolute = array_all_names[damerau_levenshtein_distance_ndarray(term, array_all_names) <= 2]
            array_absolute_values = damerau_levenshtein_distance_ndarray(term, array_absolute)
            if bln_print:
                print(array_absolute, array_absolute_values)
            set_intersection = set(array_normalized).intersection(set(array_absolute))
            # check if a similar name was found
            term_approx = ''
            if len(set_intersection) == 1:
                term_approx = list(set_intersection)[0]
                if bln_print:
                    print('we found the name', set_intersection)
            elif len(set_intersection) > 1:
                # !!! we only look at normalized values
                # !!! we don't account for names with same values !!!
                array_min = array_normalized[array_normalized_values.argmin()]
                term_approx = array_min#[0]
                if bln_print:
                    print('we found several possible names', set_intersection, 'and choose', array_min)
            if term_approx:
                str_name, str_role, int_uniqueID = get_string(term_approx, str_name, str_role, int_uniqueID, str_canton)
                print('*******************', str_name, term_approx)


    return str_name, str_role, int_uniqueID, find_names



# function to get data frame from lists of names
# input:
# - lists_names: lists of names (simple, double, comp, canton)
# output:
# - df: corresponding dataframe
def get_df_from_lists_names(lists_names):
    list_types = ['simple', 'double', 'comp', 'canton']
    df = pd.DataFrame()
    for i in range(4):
        df_temp = pd.DataFrame(lists_names[i],
                columns = ('name_short', 'name_correct', 'uniqueIndex', 'CantonName', 'CantonAbbreviation', 'Citizenship', 'FirstName'))
        df_temp['type'] = list_types[i]
        df = pd.concat([df, df_temp], ignore_index = True)
    return df




# function to extract votation paragraphs
# !!! maybe we only need a list of votation paragraphs
# !!! error prone, possible improvements see notebook extract_discussions
# input:
# - dict_text: dictionary with text of one file
# - list_names: list of votation terms
# - bln_print: whether to print during execution, default False
# output:
# - dict_votations: dictionary with votations
def get_votations(dict_text, list_votationterms, bln_print=True):
    count = 0
    dict_votations = {}
    for key, text in dict_text.items():
        list_oi = tokenizer.tokenize(text)[:15]
#        if len(set(list_oi).intersection(set(list_votationterms))) > 1:
        if set(list_oi).intersection(set(list_votationterms)):
                count += 1
                dict_votations[key] = text
                if bln_print:
                    print(count, 'MATCH', key, list_oi)
        else:
            #pass
            if bln_print:
                print('-----     ', list_oi)

    if bln_print:
        print(count)

    return dict_votations




# function to put discussions together
# !!! needs improvement when OCRed xml is corrected (details see notebook)
# input:
# - dict_discussionstarts
# - dict_votations
# output:
# - dict_discussions: dictionary of discussion parts
#                        key: integer of discussion start
#                        value: text until next discussion start or votation paragraph
def get_discussions(dict_text, dict_discussionstarts, dict_votations):

    # helper function to add text to discussion dictionary
    def add_to_dict(key, i):
#        print(key, i)
        if key not in dict_discussions:
            dict_discussions[key] = dict_discussionstarts[key]
        else:
            if i in list_text_keys_integers:
                actual_i = list(dict_text.keys())[list_text_keys_integers.index(i)]
                only_text = dict_discussions[key][1] + dict_text[actual_i]
                dict_discussions[key] = (dict_discussions[key][0], only_text)
        list_keys.append(i)


    # list of keys for discussion starts and votation paragraphs
    list_discussionstarts = list(dict_discussionstarts.keys())
    list_discussionstarts_integers = [int(tpl[0].split(',')[0]) for tpl in dict_discussionstarts.keys()]
    list_votations_strings = list(dict_votations.keys())
    list_votations_integers = [int(tpl.split(',')[0]) for tpl in list_votations_strings]
    list_text_keys_integers = [int(tpl.split(',')[0]) for tpl in dict_text.keys()]

    # initialize empty dictionary for discussions and empty list for all added keys
    dict_discussions = {}
    list_keys = []

    # if there are no discussion starts, return empty dictionary and list
    if not list_discussionstarts:
        return dict_discussions, list_keys

    # for every discussion start except last
    for idx, key in enumerate(list_discussionstarts_integers[:-1]):
        #print(idx, key)
        # write discussion start to dictionary
        add_to_dict(list_discussionstarts[idx], key)

        # for every textbox until next discussion start
        for i in range(key + 1, list_discussionstarts_integers[idx + 1]):
            # if it is not a votation paragraph, write it to dictionary,
            if i not in list_votations_integers:
                add_to_dict(list_discussionstarts[idx], i)
            # else, stop execution of for loop
            else:
                break

    # for last discussion start
    last_key = list_discussionstarts_integers[-1]
    # write discussion start to dictionary
    add_to_dict(list_discussionstarts[-1], last_key)
    # for every textbox until the end of the document
    for i in range(last_key + 1, max(list_text_keys_integers) + 1):
        # if it is not a votation paragraph, write it to dictionary
        if i not in list_votations_strings:
            add_to_dict(list_discussionstarts[-1], i)
        # else, stop execution of for loop
        else:
            break

    return dict_discussions, list_keys


# function to check whether a file containts discussions
# achieved by excluding title pages, table of content, etc.
# !!! function works well for 1891 - 1900, not checked after that !!!
def check_if_discussion(path_meta_xml_file,
        list_attributes  = ['TITEL_NORMAL_DE', 'TITEL_NORMAL_FR', 'TITEL_ORIGINAL_DE', 'TITEL_ORIGINAL_FR'],
        list_nondiscussion = ['inhaltsverzeichnis', 'inhaltsverzeichniss', 'jahresinhalt', 'rednerliste',
            'umschlag', 'sachregister', 'titelblatt', 'numerierung'],
        list_nondiscussion2 = ['table', 'matières', 'répertoire', 'procès-verbaux']):
    # parse, get root and then part of interest
    XML_tree = ET.parse(path_meta_xml_file)
    XML_root = XML_tree.getroot()
    XML_poi = XML_root[0].find('ADS_TEXTEINHEIT')

    # for each title attribute
    for attribute in list_attributes:
        # if xml contains this attribute
        if attribute in XML_poi.attrib:
            # get title and generate set with lower case terms
            title = XML_poi.attrib[attribute]
            set_title = set([term.lower() for term in title.split()])
            #print(set_title)
            # if one of terms is in list_nondiscussion, return False
            if set_title.intersection(set(list_nondiscussion)):
                #print('NOOO', path_meta_xml_file)
                return False
            # if two terms are in list_nondiscussion2, also return False
            if len(set_title.intersection(set(list_nondiscussion2))) > 1:
                #print('NOOO', path_meta_xml_file)
                return False

    return True





def tokenize_dictionary(dictionary, tokenizer, only_text=False):
    dictionary_tokenized = {}
    # if there is only text, e.g. when we look at all texts of a document at once (level 2 in flattened dictionary)
    if only_text:
        for key, text in dictionary.items():
            dictionary_tokenized[key] = tokenizer.tokenize(text)
    # if the values are actually tuples (speaker, text), e.g. when a document corresponds to what one person said
    else:
        for key, text in dictionary.items():
            dictionary_tokenized[key] = (text[0], tokenizer.tokenize(text[1]))

    return dictionary_tokenized



# value of dictionary needs to be tokenized !!!!
def remove_stopwords_from_dictionary(dictionary, list_stopwords, only_text=False):
    dict_docs_afterswr = {}
    # if there is only text, e.g. when we look at all texts of a document at once (level 2 in flattened dictionary)
    if only_text:
        for doc, text in dictionary.items():
            list_text = text
            list_words_tokenized = [word for word in list_text if word.lower() not in list_stopwords]
            dict_docs_afterswr[doc] = ' '.join(list_words_tokenized)
    # if the values are actually tuples (speaker, text), e.g. when a document corresponds to what one person said
    else:
        for doc, tupel in dictionary.items():
            before_colon = tupel[0]
            list_text = tupel[1]
            list_words_tokenized = [word for word in list_text if word.lower() not in list_stopwords]
            dict_docs_afterswr[doc] = (before_colon, ' '.join(list_words_tokenized))

    return dict_docs_afterswr



def dict_only_text(dictionary):
    dictionary_only_text = {}
    for key, tupel in dictionary.items():
        dictionary_only_text[key] = tupel[1]

    return dictionary_only_text




# two functions for language identification
# Author: Luis Salamanca
# small modifications by Lili Gasser
# Using stopwords
def identify_lang(dict_text, valid_lang = ('german', 'french', 'italian')):

    language_ratios_textbox = {}

    for i_k in dict_text.keys():
        tokens = dict_text[i_k][1]
        test_words = [word.lower() for word in tokens] # lowercase all tokens
        test_words_set = set(test_words)
        language_ratios = {}
        for language in stopwords.fileids():
            if language in valid_lang:
                stopwords_set = set(stopwords.words(language)) # For some languages eg. Russian, it would be a wise idea to tokenize the stop words by punctuation too.
                common_elements = test_words_set.intersection(stopwords_set)
                language_ratios[language] = len(common_elements) # language "score"
        language_ratios_textbox[i_k] = language_ratios

    return language_ratios_textbox


# Simply, given the number of ocurrences of the stopwords, it assigns a label
# to a specific textbox, also considering the possibility of textboxes
# mixing languages. For this case, the value ratio_similar is intended

def label_language(aux_dict_l):
    ratio_similar = 0.8
    if sum(aux_dict_l.values()):
        aux_dict_l_norm = {k: v / total for total in (sum(aux_dict_l.values()),) for k, v in aux_dict_l.items()}
        lang_max_aux = max(aux_dict_l_norm.keys(), key=(lambda key: aux_dict_l_norm[key]))
        lang_max = ''
        count_l = 0
        for lang in aux_dict_l_norm.keys():
            if (aux_dict_l_norm[lang] >  aux_dict_l_norm[lang_max_aux] * ratio_similar):
                if count_l > 0:
                    lang_max += '_'
                lang_max += lang
                count_l += 1
        if count_l > 1:
            lang_max = 'mixed_' + lang_max
    else:
        lang_max = 'NotIdentified'
    return lang_max


# function to get list of places
def get_list_cantons(df_names):
    df_temp = df_names.loc[df_names['type']=='canton']
    list_cantonname = list(df_temp['CantonName'])
    list_cantonabbr = list(df_temp['CantonAbbreviation'])
    list_citizenship = list(df_temp['Citizenship'])
    list_citizenship = [city[:-5] for item in list_citizenship for city in item.split(',')]
    list_firstname = list(df_temp['FirstName'])

    return list_cantonname +  list_cantonabbr + list_citizenship + list_firstname
